import gc
import sys
import unittest
import warnings

from contextlib import contextmanager
from functools import partial

from igraph import (
    ALL,
    Edge,
    EdgeSeq,
    Graph,
    IN,
    InternalError,
    is_degree_sequence,
    is_graphical,
    is_graphical_degree_sequence,
    Matrix,
    Vertex,
    VertexSeq,
)
from igraph._igraph import EdgeSeq as _EdgeSeq, VertexSeq as _VertexSeq

from .utils import is_pypy

try:
    import numpy as np
except ImportError:
    np = None


class BasicTests(unittest.TestCase):
    def testGraphCreation(self):
        g = Graph()
        self.assertTrue(isinstance(g, Graph))
        self.assertTrue(g.vcount() == 0 and g.ecount() == 0 and not g.is_directed())

        g = Graph(3, [(0, 1), (1, 2), (2, 0)])
        self.assertTrue(
            g.vcount() == 3
            and g.ecount() == 3
            and not g.is_directed()
            and g.is_simple()
        )

        g = Graph(2, [(0, 1), (1, 2), (2, 3)], True)
        self.assertTrue(
            g.vcount() == 4 and g.ecount() == 3 and g.is_directed() and g.is_simple()
        )

        g = Graph([(0, 1), (1, 2), (2, 1)])
        self.assertTrue(
            g.vcount() == 3
            and g.ecount() == 3
            and not g.is_directed()
            and not g.is_simple()
        )

        g = Graph(((0, 1), (0, 0), (1, 2)))
        self.assertTrue(
            g.vcount() == 3
            and g.ecount() == 3
            and not g.is_directed()
            and not g.is_simple()
        )

        g = Graph(8, None)
        self.assertEqual(8, g.vcount())
        self.assertEqual(0, g.ecount())
        self.assertFalse(g.is_directed())

        g = Graph(edges=None)
        self.assertEqual(0, g.vcount())
        self.assertEqual(0, g.ecount())
        self.assertFalse(g.is_directed())

        self.assertRaises(TypeError, Graph, edgelist=[(1, 2)])

    @unittest.skipIf(np is None, "test case depends on NumPy")
    def testGraphCreationWithNumPy(self):
        # NumPy array with integers
        arr = np.array([(0, 1), (1, 2), (2, 3)])
        g = Graph(arr, directed=True)
        self.assertTrue(
            g.vcount() == 4 and g.ecount() == 3 and g.is_directed() and g.is_simple()
        )

        # Sliced NumPy array -- the sliced array is non-contiguous but we
        # automatically make it so
        arr = np.array([(0, 1), (10, 11), (1, 2), (11, 12), (2, 3), (12, 13)])
        g = Graph(arr[::2, :], directed=True)
        self.assertTrue(
            g.vcount() == 4 and g.ecount() == 3 and g.is_directed() and g.is_simple()
        )

        # 1D NumPy array -- should raise a TypeError because we need a 2D array
        arr = np.array([0, 1, 1, 2, 2, 3])
        self.assertRaises(TypeError, Graph, arr)

        # 3D NumPy array -- should raise a TypeError because we need a 2D array
        arr = np.array([([0, 1], [10, 11]), ([1, 2], [11, 12]), ([2, 3], [12, 13])])
        self.assertRaises(TypeError, Graph, arr)

        # NumPy array with strings -- should be a casting error
        arr = np.array([("a", "b"), ("c", "d"), ("e", "f")])
        self.assertRaises(ValueError, Graph, arr)

    def testAddVertex(self):
        g = Graph()

        vertex = g.add_vertex()
        self.assertTrue(g.vcount() == 1 and g.ecount() == 0)
        self.assertEqual(0, vertex.index)
        self.assertFalse("name" in g.vertex_attributes())

        vertex = g.add_vertex("foo")
        self.assertTrue(g.vcount() == 2 and g.ecount() == 0)
        self.assertEqual(1, vertex.index)
        self.assertTrue("name" in g.vertex_attributes())
        self.assertEqual(g.vs["name"], [None, "foo"])

        vertex = g.add_vertex("3")
        self.assertTrue(g.vcount() == 3 and g.ecount() == 0)
        self.assertEqual(2, vertex.index)
        self.assertTrue("name" in g.vertex_attributes())
        self.assertEqual(g.vs["name"], [None, "foo", "3"])

        vertex = g.add_vertex(name="bar")
        self.assertTrue(g.vcount() == 4 and g.ecount() == 0)
        self.assertEqual(3, vertex.index)
        self.assertTrue("name" in g.vertex_attributes())
        self.assertEqual(g.vs["name"], [None, "foo", "3", "bar"])

        vertex = g.add_vertex(name="frob", spam="cheese", ham=42)
        self.assertTrue(g.vcount() == 5 and g.ecount() == 0)
        self.assertEqual(4, vertex.index)
        self.assertEqual(sorted(g.vertex_attributes()), ["ham", "name", "spam"])
        self.assertEqual(g.vs["spam"], [None] * 4 + ["cheese"])
        self.assertEqual(g.vs["ham"], [None] * 4 + [42])

        with self.assertWarns(DeprecationWarning, msg="integers as vertex names"):
            g.add_vertex(42)

    def testAddVertices(self):
        g = Graph()
        g.add_vertices(2)
        self.assertTrue(g.vcount() == 2 and g.ecount() == 0)

        g.add_vertices("spam")
        self.assertTrue(g.vcount() == 3 and g.ecount() == 0)
        self.assertEqual(g.vs[2]["name"], "spam")

        g.add_vertices(["bacon", "eggs"])
        self.assertTrue(g.vcount() == 5 and g.ecount() == 0)
        self.assertEqual(g.vs[2:]["name"], ["spam", "bacon", "eggs"])

        g.add_vertices(2, attributes={"color": ["k", "b"]})
        self.assertEqual(g.vs[2:]["name"], ["spam", "bacon", "eggs", None, None])
        self.assertEqual(g.vs[5:]["color"], ["k", "b"])

    def testDeleteVertices(self):
        g = Graph([(0, 1), (1, 2), (2, 3), (0, 2), (3, 4), (4, 5)])
        self.assertEqual(6, g.vcount())
        self.assertEqual(6, g.ecount())

        # Delete a single vertex
        g.delete_vertices(4)
        self.assertEqual(5, g.vcount())
        self.assertEqual(4, g.ecount())

        # Delete multiple vertices
        g.delete_vertices([1, 3])
        self.assertEqual(3, g.vcount())
        self.assertEqual(1, g.ecount())

        # Delete a vertex sequence
        g.delete_vertices(g.vs[:2])
        self.assertEqual(1, g.vcount())
        self.assertEqual(0, g.ecount())

        # Delete a single vertex object
        g.vs[0].delete()
        self.assertEqual(0, g.vcount())
        self.assertEqual(0, g.ecount())

        # Delete vertices by name
        g = Graph.Full(4)
        g.vs["name"] = ["spam", "bacon", "eggs", "ham"]
        self.assertEqual(4, g.vcount())
        g.delete_vertices("spam")
        self.assertEqual(3, g.vcount())
        g.delete_vertices(["bacon", "ham"])
        self.assertEqual(1, g.vcount())

        # Deleting a nonexistent vertex
        self.assertRaises(ValueError, g.delete_vertices, "no-such-vertex")
        self.assertRaises(InternalError, g.delete_vertices, 2)

        # Delete all vertices
        g.delete_vertices()
        self.assertEqual(0, g.vcount())

    def testAddEdge(self):
        g = Graph()
        g.add_vertices(["spam", "bacon", "eggs", "ham"])

        edge = g.add_edge(0, 1)
        self.assertEqual(g.vcount(), 4)
        self.assertEqual(g.get_edgelist(), [(0, 1)])
        self.assertEqual(0, edge.index)
        self.assertEqual((0, 1), edge.tuple)

        edge = g.add_edge(1, 2, foo="bar")
        self.assertEqual(g.vcount(), 4)
        self.assertEqual(g.get_edgelist(), [(0, 1), (1, 2)])
        self.assertEqual(1, edge.index)
        self.assertEqual((1, 2), edge.tuple)
        self.assertEqual("bar", edge["foo"])
        self.assertEqual([None, "bar"], g.es["foo"])

    def testAddEdges(self):
        g = Graph()
        g.add_vertices(["spam", "bacon", "eggs", "ham"])

        g.add_edges([(0, 1)])
        self.assertEqual(g.vcount(), 4)
        self.assertEqual(g.get_edgelist(), [(0, 1)])

        g.add_edges([(1, 2), (2, 3), (1, 3)])
        self.assertEqual(g.vcount(), 4)
        self.assertEqual(g.get_edgelist(), [(0, 1), (1, 2), (2, 3), (1, 3)])

        g.add_edges([("spam", "eggs"), ("spam", "ham")])
        self.assertEqual(g.vcount(), 4)
        self.assertEqual(
            g.get_edgelist(), [(0, 1), (1, 2), (2, 3), (1, 3), (0, 2), (0, 3)]
        )

        g.add_edges([(0, 0), (1, 1)], attributes={"color": ["k", "b"]})
        self.assertEqual(
            g.get_edgelist(),
            [
                (0, 1),
                (1, 2),
                (2, 3),
                (1, 3),
                (0, 2),
                (0, 3),
                (0, 0),
                (1, 1),
            ],
        )
        self.assertEqual(g.es["color"], [None, None, None, None, None, None, "k", "b"])

    def testDeleteEdges(self):
        g = Graph.Famous("petersen")
        g.vs["name"] = list("ABCDEFGHIJ")
        el = g.get_edgelist()

        self.assertEqual(15, g.ecount())

        # Deleting single edge
        g.delete_edges(14)
        el[14:] = []
        self.assertEqual(14, g.ecount())
        self.assertEqual(el, g.get_edgelist())

        # Deleting multiple edges
        g.delete_edges([2, 5, 7])
        el[7:8] = []
        el[5:6] = []
        el[2:3] = []
        self.assertEqual(11, g.ecount())
        self.assertEqual(el, g.get_edgelist())

        # Deleting edge object
        g.es[6].delete()
        el[6:7] = []
        self.assertEqual(10, g.ecount())
        self.assertEqual(el, g.get_edgelist())

        # Deleting edge sequence object
        g.es[1:4].delete()
        el[1:4] = []
        self.assertEqual(7, g.ecount())
        self.assertEqual(el, g.get_edgelist())

        # Deleting edges by IDs
        g.delete_edges([(2, 7), (5, 8)])
        el[4:5] = []
        el[1:2] = []
        self.assertEqual(5, g.ecount())
        self.assertEqual(el, g.get_edgelist())

        # Deleting edges by names
        g.delete_edges([("D", "I"), ("G", "I")])
        el[3:4] = []
        el[1:2] = []
        self.assertEqual(3, g.ecount())
        self.assertEqual(el, g.get_edgelist())

        # Deleting nonexistent edges
        self.assertRaises(ValueError, g.delete_edges, [(0, 2)])
        self.assertRaises(ValueError, g.delete_edges, [("A", "C")])
        self.assertRaises(ValueError, g.delete_edges, [(0, 15)])

        # Delete all edges
        g.delete_edges()
        self.assertEqual(0, g.ecount())

    def testClear(self):
        g = Graph.Famous("petersen")
        g["name"] = list("petersen")

        # Clearing the graph
        g.clear()

        self.assertEqual(0, g.vcount())
        self.assertEqual(0, g.ecount())
        self.assertEqual([], g.attributes())

    def testGraphGetEid(self):
        g = Graph.Famous("petersen")
        g.vs["name"] = list("ABCDEFGHIJ")
        edges_to_ids = {v: k for k, v in enumerate(g.get_edgelist())}
        for (source, target), edge_id in edges_to_ids.items():
            source_name, target_name = g.vs[(source, target)]["name"]
            self.assertEqual(edge_id, g.get_eid(source, target))
            self.assertEqual(edge_id, g.get_eid(source_name, target_name))

        self.assertRaises(InternalError, g.get_eid, 0, 11)
        self.assertRaises(ValueError, g.get_eid, "A", "K")

    def testGraphGetEids(self):
        g = Graph.Famous("petersen")
        eids = g.get_eids(pairs=[(0, 1), (0, 5), (1, 6), (4, 9), (8, 6)])
        self.assertTrue(eids == [0, 2, 4, 9, 12])
        eids = g.get_eids(pairs=[(7, 9), (9, 6)])
        self.assertTrue(eids == [14, 13])
        self.assertRaises(InternalError, g.get_eids, pairs=[(0, 1), (0, 2)])
        self.assertRaises(TypeError, g.get_eids, pairs=None)

    def testAdjacency(self):
        g = Graph(4, [(0, 1), (1, 2), (2, 0), (2, 3)], directed=True)
        self.assertTrue(g.neighbors(2) == [0, 1, 3])
        self.assertTrue(g.predecessors(2) == [1])
        self.assertTrue(g.successors(2) == [0, 3])
        self.assertTrue(g.get_adjlist() == [[1], [2], [0, 3], []])
        self.assertTrue(g.get_adjlist(IN) == [[2], [0], [1], [2]])
        self.assertTrue(g.get_adjlist(ALL) == [[1, 2], [0, 2], [0, 1, 3], [2]])

    def testEdgeIncidence(self):
        g = Graph(4, [(0, 1), (1, 2), (2, 0), (2, 3)], directed=True)
        self.assertTrue(g.incident(2) == [2, 3])
        self.assertTrue(g.incident(2, IN) == [1])
        self.assertTrue(g.incident(2, ALL) == [2, 1, 3])
        self.assertTrue(g.get_inclist() == [[0], [1], [2, 3], []])
        self.assertTrue(g.get_inclist(IN) == [[2], [0], [1], [3]])
        self.assertTrue(g.get_inclist(ALL) == [[0, 2], [0, 1], [2, 1, 3], [3]])

    def testMultiplesLoops(self):
        g = Graph.Tree(7, 2)

        # has_multiple
        self.assertFalse(g.has_multiple())

        g.add_vertices(1)
        g.add_edges([(0, 1), (7, 7), (6, 6), (6, 6), (6, 6)])

        # is_loop
        self.assertTrue(
            g.is_loop()
            == [False, False, False, False, False, False, False, True, True, True, True]
        )
        self.assertTrue(g.is_loop(g.ecount() - 2))
        self.assertTrue(g.is_loop(list(range(6, 8))) == [False, True])

        # is_multiple
        self.assertTrue(
            g.is_multiple()
            == [
                False,
                False,
                False,
                False,
                False,
                False,
                True,
                False,
                False,
                True,
                True,
            ]
        )

        # has_multiple
        self.assertTrue(g.has_multiple())

        # count_multiple
        self.assertTrue(g.count_multiple() == [2, 1, 1, 1, 1, 1, 2, 1, 3, 3, 3])
        self.assertTrue(g.count_multiple(g.ecount() - 1) == 3)
        self.assertTrue(g.count_multiple(list(range(2, 5))) == [1, 1, 1])

        # check if a mutual directed edge pair is reported as multiple
        g = Graph(2, [(0, 1), (1, 0)], directed=True)
        self.assertTrue(g.is_multiple() == [False, False])

    def testPickling(self):
        import pickle

        g = Graph([(0, 1), (1, 2)])
        g["data"] = "abcdef"
        g.vs["data"] = [3, 4, 5]
        g.es["data"] = ["A", "B"]
        g.custom_data = None
        pickled = pickle.dumps(g)

        g2 = pickle.loads(pickled)
        self.assertTrue(g["data"] == g2["data"])
        self.assertTrue(g.vs["data"] == g2.vs["data"])
        self.assertTrue(g.es["data"] == g2.es["data"])
        self.assertTrue(g.vcount() == g2.vcount())
        self.assertTrue(g.ecount() == g2.ecount())
        self.assertTrue(g.is_directed() == g2.is_directed())
        self.assertTrue(g2.custom_data == g.custom_data)

    def testHashing(self):
        g = Graph([(0, 1), (1, 2)])
        self.assertRaises(TypeError, hash, g)

    def testIteration(self):
        g = Graph()
        self.assertRaises(TypeError, iter, g)


class DatatypeTests(unittest.TestCase):
    def testMatrix(self):
        m = Matrix([[1, 2, 3], [4, 5], [6, 7, 8]])
        self.assertTrue(m.shape == (3, 3))

        # Reading data
        self.assertTrue(m.data == [[1, 2, 3], [4, 5, 0], [6, 7, 8]])
        self.assertTrue(m[1, 1] == 5)
        self.assertTrue(m[0] == [1, 2, 3])
        self.assertTrue(m[0, :] == [1, 2, 3])
        self.assertTrue(m[:, 0] == [1, 4, 6])
        self.assertTrue(m[2, 0:2] == [6, 7])
        self.assertTrue(m[:, :].data == [[1, 2, 3], [4, 5, 0], [6, 7, 8]])
        self.assertTrue(m[:, 1:3].data == [[2, 3], [5, 0], [7, 8]])

        # Writing data
        m[1, 1] = 10
        self.assertTrue(m[1, 1] == 10)
        m[1] = (6, 5, 4)
        self.assertTrue(m[1] == [6, 5, 4])
        m[1:3] = [[4, 5, 6], (7, 8, 9)]
        self.assertTrue(m[1:3].data == [[4, 5, 6], [7, 8, 9]])

        # Minimums and maximums
        self.assertTrue(m.min() == 1)
        self.assertTrue(m.max() == 9)
        self.assertTrue(m.min(0) == [1, 2, 3])
        self.assertTrue(m.max(0) == [7, 8, 9])
        self.assertTrue(m.min(1) == [1, 4, 7])
        self.assertTrue(m.max(1) == [3, 6, 9])

        # Special constructors
        m = Matrix.Fill(2, (3, 3))
        self.assertTrue(m.min() == 2 and m.max() == 2 and m.shape == (3, 3))
        m = Matrix.Zero(5, 4)
        self.assertTrue(m.min() == 0 and m.max() == 0 and m.shape == (5, 4))
        m = Matrix.Identity(3)
        self.assertTrue(m.data == [[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        m = Matrix.Identity(3, 2)
        self.assertTrue(m.data == [[1, 0], [0, 1], [0, 0]])

        # Conversion to string
        m = Matrix.Identity(3)
        self.assertTrue(str(m) == "[[1, 0, 0]\n [0, 1, 0]\n [0, 0, 1]]")
        self.assertTrue(repr(m) == "Matrix([[1, 0, 0], [0, 1, 0], [0, 0, 1]])")


class GraphDictListTests(unittest.TestCase):
    def setUp(self):
        self.vertices = [
            {"name": "Alice", "age": 48, "gender": "F"},
            {"name": "Bob", "age": 33, "gender": "M"},
            {"name": "Cecil", "age": 45, "gender": "F"},
            {"name": "David", "age": 34, "gender": "M"},
        ]
        self.edges = [
            {"source": "Alice", "target": "Bob", "friendship": 4, "advice": 4},
            {"source": "Cecil", "target": "Bob", "friendship": 5, "advice": 5},
            {"source": "Cecil", "target": "Alice", "friendship": 5, "advice": 5},
            {"source": "David", "target": "Alice", "friendship": 2, "advice": 4},
            {"source": "David", "target": "Bob", "friendship": 1, "advice": 2},
        ]

    def testGraphFromDictList(self):
        g = Graph.DictList(self.vertices, self.edges)
        self.checkIfOK(g, "name")
        g = Graph.DictList(self.vertices, self.edges, iterative=True)
        self.checkIfOK(g, "name")

    def testGraphFromDictIterator(self):
        g = Graph.DictList(iter(self.vertices), iter(self.edges))
        self.checkIfOK(g, "name")
        g = Graph.DictList(iter(self.vertices), iter(self.edges), iterative=True)
        self.checkIfOK(g, "name")

    def testGraphFromDictIteratorNoVertices(self):
        g = Graph.DictList(None, iter(self.edges))
        self.checkIfOK(g, "name", check_vertex_attrs=False)
        g = Graph.DictList(None, iter(self.edges), iterative=True)
        self.checkIfOK(g, "name", check_vertex_attrs=False)

    def testGraphFromDictListExtraVertexName(self):
        del self.vertices[2:]  # No data for "Cecil" and "David"
        g = Graph.DictList(self.vertices, self.edges)
        self.assertTrue(g.vcount() == 4 and g.ecount() == 5 and not g.is_directed())
        self.assertTrue(g.vs["name"] == ["Alice", "Bob", "Cecil", "David"])
        self.assertTrue(g.vs["age"] == [48, 33, None, None])
        self.assertTrue(g.vs["gender"] == ["F", "M", None, None])
        self.assertTrue(g.es["friendship"] == [4, 5, 5, 2, 1])
        self.assertTrue(g.es["advice"] == [4, 5, 5, 4, 2])
        self.assertTrue(g.get_edgelist() == [(0, 1), (1, 2), (0, 2), (0, 3), (1, 3)])

    def testGraphFromDictListAlternativeName(self):
        for vdata in self.vertices:
            vdata["name_alternative"] = vdata["name"]
            del vdata["name"]
        g = Graph.DictList(
            self.vertices, self.edges, vertex_name_attr="name_alternative"
        )
        self.checkIfOK(g, "name_alternative")
        g = Graph.DictList(
            self.vertices,
            self.edges,
            vertex_name_attr="name_alternative",
            iterative=True,
        )
        self.checkIfOK(g, "name_alternative")

    def checkIfOK(self, g, name_attr, check_vertex_attrs=True):
        self.assertTrue(g.vcount() == 4 and g.ecount() == 5 and not g.is_directed())
        self.assertTrue(g.get_edgelist() == [(0, 1), (1, 2), (0, 2), (0, 3), (1, 3)])
        self.assertTrue(g.vs[name_attr] == ["Alice", "Bob", "Cecil", "David"])
        if check_vertex_attrs:
            self.assertTrue(g.vs["age"] == [48, 33, 45, 34])
            self.assertTrue(g.vs["gender"] == ["F", "M", "F", "M"])
        self.assertTrue(g.es["friendship"] == [4, 5, 5, 2, 1])
        self.assertTrue(g.es["advice"] == [4, 5, 5, 4, 2])


class GraphTupleListTests(unittest.TestCase):
    def setUp(self):
        self.edges = [
            ("Alice", "Bob", 4, 4),
            ("Cecil", "Bob", 5, 5),
            ("Cecil", "Alice", 5, 5),
            ("David", "Alice", 2, 4),
            ("David", "Bob", 1, 2),
        ]

    def testGraphFromTupleList(self):
        g = Graph.TupleList(self.edges)
        self.checkIfOK(g, "name", ())

    def testGraphFromTupleListWithEdgeAttributes(self):
        g = Graph.TupleList(self.edges, edge_attrs=("friendship", "advice"))
        self.checkIfOK(g, "name", ("friendship", "advice"))
        g = Graph.TupleList(self.edges, edge_attrs=("friendship",))
        self.checkIfOK(g, "name", ("friendship",))
        g = Graph.TupleList(self.edges, edge_attrs="friendship")
        self.checkIfOK(g, "name", ("friendship",))

    def testGraphFromTupleListWithDifferentNameAttribute(self):
        g = Graph.TupleList(self.edges, vertex_name_attr="spam")
        self.checkIfOK(g, "spam", ())

    def testGraphFromTupleListWithWeights(self):
        g = Graph.TupleList(self.edges, weights=True)
        self.checkIfOK(g, "name", ("weight",))
        g = Graph.TupleList(self.edges, weights="friendship")
        self.checkIfOK(g, "name", ("friendship",))
        g = Graph.TupleList(self.edges, weights=False)
        self.checkIfOK(g, "name", ())
        self.assertRaises(
            ValueError,
            Graph.TupleList,
            [self.edges],
            weights=True,
            edge_attrs="friendship",
        )

    def testNoneForMissingAttributes(self):
        g = Graph.TupleList(self.edges, edge_attrs=("friendship", "advice", "spam"))
        self.checkIfOK(g, "name", ("friendship", "advice", "spam"))

    def checkIfOK(self, g, name_attr, edge_attrs):
        self.assertTrue(g.vcount() == 4 and g.ecount() == 5 and not g.is_directed())
        self.assertTrue(g.get_edgelist() == [(0, 1), (1, 2), (0, 2), (0, 3), (1, 3)])
        self.assertTrue(g.attributes() == [])
        self.assertTrue(g.vertex_attributes() == [name_attr])
        self.assertTrue(g.vs[name_attr] == ["Alice", "Bob", "Cecil", "David"])
        if edge_attrs:
            self.assertTrue(sorted(g.edge_attributes()) == sorted(edge_attrs))
            self.assertTrue(g.es[edge_attrs[0]] == [4, 5, 5, 2, 1])
            if len(edge_attrs) > 1:
                self.assertTrue(g.es[edge_attrs[1]] == [4, 5, 5, 4, 2])
            if len(edge_attrs) > 2:
                self.assertTrue(g.es[edge_attrs[2]] == [None] * 5)
        else:
            self.assertTrue(g.edge_attributes() == [])


class GraphListDictTests(unittest.TestCase):
    def setUp(self):
        self.eids = {
            0: [1],
            2: [1, 0],
            3: [0, 1],
        }
        self.edges = {
            "Alice": ["Bob"],
            "Cecil": ["Bob", "Alice"],
            "David": ["Alice", "Bob"],
        }

    def testEmptyGraphListDict(self):
        g = Graph.ListDict({})
        self.assertEqual(g.vcount(), 0)

    def testGraphFromListDict(self):
        g = Graph.ListDict(self.eids)
        self.checkIfOK(g, ())

    def testGraphFromListDictWithNames(self):
        g = Graph.ListDict(self.edges)
        self.checkIfOK(g, "name")

    def checkIfOK(self, g, name_attr):
        self.assertTrue(g.vcount() == 4 and g.ecount() == 5 and not g.is_directed())
        self.assertTrue(g.get_edgelist() == [(0, 1), (1, 2), (0, 2), (0, 3), (1, 3)])
        self.assertTrue(g.attributes() == [])
        if name_attr:
            self.assertTrue(g.vertex_attributes() == [name_attr])
            self.assertTrue(g.vs[name_attr] == ["Alice", "Bob", "Cecil", "David"])
        self.assertTrue(g.edge_attributes() == [])


class GraphDictDictTests(unittest.TestCase):
    def setUp(self):
        self.eids = {
            0: {1: {}},
            2: {1: {}, 0: {}},
            3: {0: {}, 1: {}},
        }
        self.edges = {
            "Alice": {"Bob": {}},
            "Cecil": {"Bob": {}, "Alice": {}},
            "David": {"Alice": {}, "Bob": {}},
        }
        self.eids_with_props = {
            0: {1: {"weight": 5.6, "additional": "abc"}},
            2: {1: {"weight": 3.4}, 0: {"weight": 2}},
            3: {0: {"weight": 1}, 1: {"weight": 5.6}},
        }

    def testEmptyGraphDictDict(self):
        g = Graph.DictDict({})
        self.assertEqual(g.vcount(), 0)

    def testGraphFromDictDict(self):
        g = Graph.DictDict(self.eids)
        self.checkIfOK(g, ())

    def testGraphFromDictDictWithProps(self):
        g = Graph.DictDict(self.eids_with_props)
        self.checkIfOK(g, (), edge_attrs=["additional", "weight"])

    def testGraphFromDictDictWithNames(self):
        g = Graph.DictDict(self.edges)
        self.checkIfOK(g, "name")

    def checkIfOK(self, g, name_attr, edge_attrs=None):
        self.assertTrue(g.vcount() == 4 and g.ecount() == 5 and not g.is_directed())
        self.assertTrue(g.get_edgelist() == [(0, 1), (1, 2), (0, 2), (0, 3), (1, 3)])
        self.assertTrue(g.attributes() == [])
        if name_attr:
            self.assertTrue(g.vertex_attributes() == [name_attr])
            self.assertTrue(g.vs[name_attr] == ["Alice", "Bob", "Cecil", "David"])
        if edge_attrs is None:
            self.assertEqual(g.edge_attributes(), [])
        else:
            self.assertEqual(sorted(g.edge_attributes()), sorted(edge_attrs))


class DegreeSequenceTests(unittest.TestCase):
    def testIsDegreeSequence(self):
        # Catch and suppress warnings because is_degree_sequence() is now
        # deprecated
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.assertTrue(is_degree_sequence([]))
            self.assertTrue(is_degree_sequence([], []))
            self.assertTrue(is_degree_sequence([0]))
            self.assertTrue(is_degree_sequence([0], [0]))
            self.assertFalse(is_degree_sequence([1]))
            self.assertTrue(is_degree_sequence([1], [1]))
            self.assertTrue(is_degree_sequence([2]))
            self.assertFalse(is_degree_sequence([2, 1, 1, 1]))
            self.assertTrue(is_degree_sequence([2, 1, 1, 1], [1, 1, 1, 2]))
            self.assertFalse(is_degree_sequence([2, 1, -2]))
            self.assertFalse(is_degree_sequence([2, 1, 1, 1], [1, 1, 1, -2]))
            self.assertTrue(is_degree_sequence([3, 3, 3, 3, 3, 3, 3, 3, 3, 3]))
            self.assertTrue(is_degree_sequence([3, 3, 3, 3, 3, 3, 3, 3, 3, 3], None))
            self.assertFalse(is_degree_sequence([3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3]))
            self.assertTrue(
                is_degree_sequence(
                    [3, 3, 3, 3, 3, 3, 3, 3, 3, 3], [4, 3, 2, 3, 4, 4, 2, 2, 4, 2]
                )
            )

    def testIsGraphicalSequence(self):
        # Catch and suppress warnings because is_graphical_degree_sequence() is now
        # deprecated
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.assertTrue(is_graphical_degree_sequence([]))
            self.assertTrue(is_graphical_degree_sequence([], []))
            self.assertTrue(is_graphical_degree_sequence([0]))
            self.assertTrue(is_graphical_degree_sequence([0], [0]))
            self.assertFalse(is_graphical_degree_sequence([1]))
            self.assertFalse(is_graphical_degree_sequence([1], [1]))
            self.assertFalse(is_graphical_degree_sequence([2]))
            self.assertFalse(is_graphical_degree_sequence([2, 1, 1, 1]))
            self.assertTrue(is_graphical_degree_sequence([2, 1, 1, 1], [1, 1, 1, 2]))
            self.assertFalse(is_graphical_degree_sequence([2, 1, -2]))
            self.assertFalse(is_graphical_degree_sequence([2, 1, 1, 1], [1, 1, 1, -2]))
            self.assertTrue(
                is_graphical_degree_sequence([3, 3, 3, 3, 3, 3, 3, 3, 3, 3])
            )
            self.assertTrue(
                is_graphical_degree_sequence([3, 3, 3, 3, 3, 3, 3, 3, 3, 3], None)
            )
            self.assertFalse(
                is_graphical_degree_sequence([3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3])
            )
            self.assertTrue(
                is_graphical_degree_sequence(
                    [3, 3, 3, 3, 3, 3, 3, 3, 3, 3], [4, 3, 2, 3, 4, 4, 2, 2, 4, 2]
                )
            )
            self.assertTrue(is_graphical_degree_sequence([3, 3, 3, 3, 4]))

    def testIsGraphicalNonSimple(self):
        # Same as testIsDegreeSequence, but using is_graphical()
        is_degree_sequence = partial(is_graphical, loops=True, multiple=True)
        self.assertTrue(is_degree_sequence([]))
        self.assertTrue(is_degree_sequence([], []))
        self.assertTrue(is_degree_sequence([0]))
        self.assertTrue(is_degree_sequence([0], [0]))
        self.assertFalse(is_degree_sequence([1]))
        self.assertTrue(is_degree_sequence([1], [1]))
        self.assertTrue(is_degree_sequence([2]))
        self.assertFalse(is_degree_sequence([2, 1, 1, 1]))
        self.assertTrue(is_degree_sequence([2, 1, 1, 1], [1, 1, 1, 2]))
        self.assertFalse(is_degree_sequence([2, 1, -2]))
        self.assertFalse(is_degree_sequence([2, 1, 1, 1], [1, 1, 1, -2]))
        self.assertTrue(is_degree_sequence([3, 3, 3, 3, 3, 3, 3, 3, 3, 3]))
        self.assertTrue(is_degree_sequence([3, 3, 3, 3, 3, 3, 3, 3, 3, 3], None))
        self.assertFalse(is_degree_sequence([3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3]))
        self.assertTrue(
            is_degree_sequence(
                [3, 3, 3, 3, 3, 3, 3, 3, 3, 3], [4, 3, 2, 3, 4, 4, 2, 2, 4, 2]
            )
        )

    def testIsGraphicalSimple(self):
        # Same as testIsGraphicalDegreeSequence, but using is_graphical()
        is_graphical_degree_sequence = partial(
            is_graphical, loops=False, multiple=False
        )
        self.assertTrue(is_graphical_degree_sequence([]))
        self.assertTrue(is_graphical_degree_sequence([], []))
        self.assertTrue(is_graphical_degree_sequence([0]))
        self.assertTrue(is_graphical_degree_sequence([0], [0]))
        self.assertFalse(is_graphical_degree_sequence([1]))
        self.assertFalse(is_graphical_degree_sequence([1], [1]))
        self.assertFalse(is_graphical_degree_sequence([2]))
        self.assertFalse(is_graphical_degree_sequence([2, 1, 1, 1]))
        self.assertTrue(is_graphical_degree_sequence([2, 1, 1, 1], [1, 1, 1, 2]))
        self.assertFalse(is_graphical_degree_sequence([2, 1, -2]))
        self.assertFalse(is_graphical_degree_sequence([2, 1, 1, 1], [1, 1, 1, -2]))
        self.assertTrue(is_graphical_degree_sequence([3, 3, 3, 3, 3, 3, 3, 3, 3, 3]))
        self.assertTrue(
            is_graphical_degree_sequence([3, 3, 3, 3, 3, 3, 3, 3, 3, 3], None)
        )
        self.assertFalse(
            is_graphical_degree_sequence([3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3])
        )
        self.assertTrue(
            is_graphical_degree_sequence(
                [3, 3, 3, 3, 3, 3, 3, 3, 3, 3], [4, 3, 2, 3, 4, 4, 2, 2, 4, 2]
            )
        )
        self.assertTrue(is_graphical_degree_sequence([3, 3, 3, 3, 4]))


class InheritedGraph(Graph):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.init_called = True

    def __new__(cls, *args, **kwds):
        result = Graph.__new__(cls)
        result.new_called = True
        return result

    @classmethod
    def Adjacency(cls, *args, **kwds):
        result = super().Adjacency(*args, **kwds)
        result.adjacency_called = True
        return result


class InheritedGraphWithEarlyMethodCallInInit(Graph):
    def __init__(self, *args, **kwds):
        self.call_result = self.degree()
        super().__init__(*args, **kwds)
        self.init_called = True


class InheritedGraphWithEarlyMethodCallInNew(Graph):
    def __new__(cls, *args, **kwds):
        instance = super().__new__(cls, *args, **kwds)
        instance.call_result = instance.degree()
        return instance


class InheritedGraphThatReturnsNonGraphFromNew(Graph):
    def __new__(cls, *args, **kwds):
        super().__new__(cls, *args, **kwds)
        return 42


class InheritanceTests(unittest.TestCase):
    def testInitCalledProperly(self):
        g = InheritedGraph()
        self.assertTrue(isinstance(g, InheritedGraph))
        self.assertTrue(getattr(g, "init_called", False))

    def testNewCalledProperly(self):
        g = InheritedGraph()
        self.assertTrue(isinstance(g, InheritedGraph))
        self.assertTrue(getattr(g, "new_called", False))

    def testInitCalledProperlyWithClassMethod(self):
        g = InheritedGraph.Tree(3, 2)
        self.assertTrue(isinstance(g, InheritedGraph))
        self.assertTrue(getattr(g, "init_called", False))

    def testNewCalledProperlyWithClassMethod(self):
        g = InheritedGraph.Tree(3, 2)
        self.assertTrue(isinstance(g, InheritedGraph))
        self.assertTrue(getattr(g, "new_called", False))

    def testCallingClassMethodInSuperclass(self):
        g = InheritedGraph.Adjacency([[0, 1, 1], [1, 0, 0], [1, 0, 0]])
        self.assertTrue(isinstance(g, InheritedGraph))
        self.assertTrue(getattr(g, "adjacency_called", True))

    def testCallingInstanceMethodTooEarly(self):
        # Creating an InheritedGraphWithEarlyMethodCallInInit instance should
        # not crash the runtime
        g = InheritedGraphWithEarlyMethodCallInInit([(0, 1), (1, 2), (2, 3)])
        self.assertEqual(4, g.vcount())
        self.assertEqual([1, 2, 2, 1], g.degree())
        self.assertEqual([], g.call_result)

        # Creating an InheritedGraphWithEarlyMethodCallInNew instance should
        # not crash the runtime either
        g = InheritedGraphWithEarlyMethodCallInNew([(0, 1), (1, 2), (2, 3)])
        self.assertEqual(4, g.vcount())
        self.assertEqual([1, 2, 2, 1], g.degree())
        self.assertEqual([], g.call_result)

    def testInheritedGraphThatReturnsSomethingElseFromNew(self):
        g = InheritedGraphThatReturnsNonGraphFromNew([(0, 1)], directed=True)
        self.assertEqual(42, g)


@contextmanager
def assert_reference_not_leaked(case, *args):
    gc.collect()
    refs_before = [sys.getrefcount(obj) for obj in args]
    try:
        yield
    finally:
        gc.collect()
        refs_after = [sys.getrefcount(obj) for obj in args]
        case.assertListEqual(refs_before, refs_after)


@unittest.skipIf(is_pypy, "reference counts are not relevant for PyPy")
class ReferenceCountTests(unittest.TestCase):
    def testEdgeReferenceCounting(self):
        with assert_reference_not_leaked(self, Edge, EdgeSeq, _EdgeSeq):
            g = Graph.Tree(3, 2)
            edge = g.es[1]
            del edge, g

    def testEdgeSeqReferenceCounting(self):
        with assert_reference_not_leaked(self, Edge, EdgeSeq, _EdgeSeq):
            g = Graph.Tree(3, 2)
            es = g.es
            es2 = EdgeSeq(g)
            del es, es2, g

    def testGraphReferenceCounting(self):
        with assert_reference_not_leaked(self, Graph, InheritedGraph):
            g = Graph.Tree(3, 2)
            self.assertTrue(gc.is_tracked(g))
            del g

    def testInheritedGraphReferenceCounting(self):
        with assert_reference_not_leaked(self, Graph, InheritedGraph):
            g = InheritedGraph.Tree(3, 2)
            self.assertTrue(gc.is_tracked(g))
            del g

    def testVertexReferenceCounting(self):
        with assert_reference_not_leaked(self, Vertex, VertexSeq, _VertexSeq):
            g = Graph.Tree(3, 2)
            vertex = g.vs[2]
            del vertex, g

    def testVertexSeqReferenceCounting(self):
        with assert_reference_not_leaked(self, Vertex, VertexSeq, _VertexSeq):
            g = Graph.Tree(3, 2)
            vs = g.vs
            vs2 = VertexSeq(g)
            del vs2, vs, g


def suite():
    basic_suite = unittest.defaultTestLoader.loadTestsFromTestCase(BasicTests)
    datatype_suite = unittest.defaultTestLoader.loadTestsFromTestCase(DatatypeTests)
    graph_dict_list_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        GraphDictListTests
    )
    graph_tuple_list_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        GraphTupleListTests
    )
    graph_list_dict_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        GraphListDictTests
    )
    graph_dict_dict_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        GraphDictDictTests
    )
    degree_sequence_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        DegreeSequenceTests
    )
    inheritance_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        InheritanceTests
    )
    refcount_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        ReferenceCountTests
    )
    return unittest.TestSuite(
        [
            basic_suite,
            datatype_suite,
            graph_dict_list_suite,
            graph_tuple_list_suite,
            graph_list_dict_suite,
            graph_dict_dict_suite,
            degree_sequence_suite,
            inheritance_suite,
            refcount_suite,
        ]
    )


def test():
    runner = unittest.TextTestRunner()
    runner.run(suite())


if __name__ == "__main__":
    test()



"""
Drawing routines to draw graphs.

This module contains routines to draw graphs on plotly surfaces.

"""

from warnings import warn

from igraph._igraph import convex_hull, VertexSeq
from igraph.drawing.baseclasses import AbstractGraphDrawer
from igraph.drawing.utils import Point

from .edge import PlotlyEdgeDrawer
from .polygon import PlotlyPolygonDrawer
from .utils import find_plotly, format_rgba
from .vertex import PlotlyVerticesDrawer

__all__ = ("PlotlyGraphDrawer",)

plotly = find_plotly()

#####################################################################


class PlotlyGraphDrawer(AbstractGraphDrawer):
    """Graph drawer that uses a pyplot.Axes as context"""

    # These need conversions, plus default passthrough for arbitrary
    # plotly shapes
    _shape_dict = {
        "rectangle": "square",
        "hidden": "none",
    }

    def __init__(
        self,
        fig,
        vertex_drawer_factory=PlotlyVerticesDrawer,
        edge_drawer_factory=PlotlyEdgeDrawer,
    ):
        """Constructs the graph drawer and associates it with the plotly Figure

        @param fig: the plotly.graph_objects.Figure to draw into.

        """
        self.fig = fig
        self.vertex_drawer_factory = vertex_drawer_factory
        self.edge_drawer_factory = edge_drawer_factory

    def draw(self, graph, *args, **kwds):
        # Deferred import to avoid a cycle in the import graph
        from igraph.clustering import VertexClustering, VertexCover

        # Positional arguments are not used
        if args:
            warn(
                "Positional arguments to plot functions are ignored "
                "and will be deprecated soon.",
                DeprecationWarning,
                stacklevel=1,
            )

        # Some abbreviations for sake of simplicity
        directed = graph.is_directed()
        fig = self.fig

        # Palette
        palette = kwds.pop("palette", None)

        # Calculate/get the layout of the graph
        layout = self.ensure_layout(kwds.get("layout", None), graph)

        # Decide whether we need to calculate the curvature of edges
        # automatically -- and calculate them if needed.
        autocurve = kwds.get("autocurve", None)
        if autocurve or (
            autocurve is None
            and "edge_curved" not in kwds
            and "curved" not in graph.edge_attributes()
            and graph.ecount() < 10000
        ):
            from igraph import autocurve

            default = kwds.get("edge_curved", 0)
            if default is True:
                default = 0.5
            default = float(default)
            kwds["edge_curved"] = autocurve(
                graph,
                attribute=None,
                default=default,
            )

        # Construct the vertex, edge and label drawers
        vertex_drawer = self.vertex_drawer_factory(fig, palette, layout)
        edge_drawer = self.edge_drawer_factory(fig, palette)

        # Construct the visual edge builders based on the specifications
        # provided by the edge_drawer
        vertex_builder = vertex_drawer.VisualVertexBuilder(graph.vs, kwds)
        edge_builder = edge_drawer.VisualEdgeBuilder(graph.es, kwds)

        # Draw the highlighted groups (if any)
        if "mark_groups" in kwds:
            mark_groups = kwds["mark_groups"]

            # Figure out what to do with mark_groups in order to be able to
            # iterate over it and get memberlist-color pairs
            if isinstance(mark_groups, dict):
                # Dictionary mapping vertex indices or tuples of vertex
                # indices to colors
                group_iter = iter(mark_groups.items())
            elif isinstance(mark_groups, (VertexClustering, VertexCover)):
                # Vertex clustering
                group_iter = ((group, color) for color, group in enumerate(mark_groups))
            elif hasattr(mark_groups, "__iter__"):
                # Lists, tuples, iterators etc
                group_iter = iter(mark_groups)
            else:
                # False
                group_iter = iter({}.items())

            # Iterate over color-memberlist pairs
            for group, color_id in group_iter:
                if not group or color_id is None:
                    continue

                color = palette.get(color_id)

                if isinstance(group, VertexSeq):
                    group = [vertex.index for vertex in group]
                if not hasattr(group, "__iter__"):
                    raise TypeError("group membership list must be iterable")

                # Get the vertex indices that constitute the convex hull
                hull = [group[i] for i in convex_hull([layout[idx] for idx in group])]

                # Calculate the preferred rounding radius for the corners
                # FIXME
                corner_radius = 1.25 * max(vertex_builder[idx].size for idx in hull)

                # Construct the polygon
                polygon = [layout[idx] for idx in hull]

                if len(polygon) == 2:
                    # Expand the polygon (which is a flat line otherwise)
                    a, b = Point(*polygon[0]), Point(*polygon[1])
                    c = corner_radius * (a - b).normalized()
                    n = Point(-c[1], c[0])
                    polygon = [a + n, b + n, b - c, b - n, a - n, a + c]
                else:
                    # Expand the polygon around its center of mass
                    center = Point(
                        *[sum(coords) / float(len(coords)) for coords in zip(*polygon)]
                    )
                    polygon = [
                        Point(*point).towards(center, -corner_radius)
                        for point in polygon
                    ]

                # Draw the hull
                facecolor = (color[0], color[1], color[2], 0.25 * color[3])
                drawer = PlotlyPolygonDrawer(fig)
                drawer.draw(
                    polygon,
                    corner_radius=corner_radius,
                    fillcolor=format_rgba(facecolor),
                    line_color=format_rgba(color),
                )

                if kwds.get("legend", False):
                    # Proxy artist for legend
                    fig.add_trace(
                        plotly.graph_objects.Bar(
                            name=str(color_id),
                            x=[],
                            y=[],
                            fillcolor=facecolor,
                            line_color=color,
                        )
                    )

            if kwds.get("legend", False):
                fig.update_layout(showlegend=True)

        # Determine the order in which we will draw the vertices and edges
        vertex_order = self._determine_vertex_order(graph, kwds)
        edge_order = self._determine_edge_order(graph, kwds)

        # Construct the iterator that we will use to draw the vertices
        vs = graph.vs
        if vertex_order is None:
            # Default vertex order
            vertex_coord_iter = zip(vs, vertex_builder, layout)
        else:
            # Specified vertex order
            vertex_coord_iter = (
                (vs[i], vertex_builder[i], layout[i]) for i in vertex_order
            )

        # Construct the iterator that we will use to draw the edges
        es = graph.es
        if edge_order is None:
            # Default edge order
            edge_coord_iter = zip(es, edge_builder)
        else:
            # Specified edge order
            edge_coord_iter = ((es[i], edge_builder[i]) for i in edge_order)

        # Draw the edges
        # We need the vertex builder to get the layout and offsets
        if directed:
            drawer_method = edge_drawer.draw_directed_edge
        else:
            drawer_method = edge_drawer.draw_undirected_edge
        for edge, visual_edge in edge_coord_iter:
            src, dest = edge.tuple
            src_vertex, dest_vertex = vertex_builder[src], vertex_builder[dest]
            drawer_method(visual_edge, src_vertex, dest_vertex)

        # Draw the vertices
        drawer_method = vertex_drawer.draw
        for vertex, visual_vertex, coords in vertex_coord_iter:
            drawer_method(visual_vertex, vertex, coords)

        # Construct the iterator that we will use to draw the vertex labels
        vs = graph.vs
        if vertex_order is None:
            # Default vertex order
            vertex_coord_iter = zip(vertex_builder, layout)
        else:
            # Specified vertex order
            vertex_coord_iter = ((vertex_builder[i], layout[i]) for i in vertex_order)

        # Draw the vertex labels
        for vertex, coords in vertex_coord_iter:
            vertex_drawer.draw_label(vertex, coords, **kwds)

        # Draw the edge labels
        labels = kwds.get("edge_label", None)
        if labels is not None:
            edge_label_iter = (
                (labels[i], edge_builder[i], graph.es[i]) for i in range(graph.ecount())
            )
            lab_args = {
                "text": [],
                "x": [],
                "y": [],
                # "textfont_color": [],
                # FIXME: horizontal/vertical alignment, offset, etc?
            }
            for label, visual_edge, edge in edge_label_iter:
                # Ask the edge drawer to propose an anchor point for the label
                src, dest = edge.tuple
                src_vertex, dest_vertex = vertex_builder[src], vertex_builder[dest]
                (x, y), (halign, valign) = edge_drawer.get_label_position(
                    visual_edge,
                    src_vertex,
                    dest_vertex,
                )
                if label is None:
                    continue

                lab_args["text"].append(label)
                lab_args["x"].append(x)
                lab_args["y"].append(y)
                # FIXME: colors do not work yet; apparently we need to convert
                # visual_edge.label_color to Plotly's format
                # lab_args["textfont_color"].append(visual_edge.label_color)
            stroke = plotly.graph_objects.Scatter(
                mode="text",
                **lab_args,
            )
            fig.add_trace(stroke)

        # Despine
        fig.update_layout(
            yaxis={"visible": False, "showticklabels": False},
            xaxis={"visible": False, "showticklabels": False},
        )



import unittest
from igraph import Graph


def assert_valid_vertex_coloring(graph, coloring):
    assert min(coloring) == 0
    for edge in graph.es:
        source, target = edge.tuple
        assert source == target or coloring[source] != coloring[target]


class VertexColoringTests(unittest.TestCase):
    def testGreedyVertexColoring(self):
        g = Graph.Famous("petersen")

        col = g.vertex_coloring_greedy()
        assert_valid_vertex_coloring(g, col)

        col = g.vertex_coloring_greedy("colored_neighbors")
        assert_valid_vertex_coloring(g, col)

        col = g.vertex_coloring_greedy("dsatur")
        assert_valid_vertex_coloring(g, col)


def suite():
    vertex_coloring_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        VertexColoringTests
    )
    return unittest.TestSuite([vertex_coloring_suite])


def test():
    runner = unittest.TextTestRunner()
    runner.run(suite())


if __name__ == "__main__":
    test()



"""Vertices drawer. Unlike other backends, all vertices are drawn at once"""

from math import pi

from igraph.drawing.baseclasses import AbstractVertexDrawer
from igraph.drawing.metamagic import AttributeCollectorBase
from .utils import find_plotly, format_rgba

__all__ = ("PlotlyVerticesDrawer",)

plotly = find_plotly()


class PlotlyVerticesDrawer(AbstractVertexDrawer):
    """Plotly backend-specific vertex drawer."""

    def __init__(self, fig, palette, layout):
        self.fig = fig
        super().__init__(palette, layout)
        self.VisualVertexBuilder = self._construct_visual_vertex_builder()

    def _construct_visual_vertex_builder(self):
        class VisualVertexBuilder(AttributeCollectorBase):
            """Collects some visual properties of a vertex for drawing"""

            _kwds_prefix = "vertex_"
            color = ("red", self.palette.get)
            frame_color = ("black", self.palette.get)
            frame_width = 1.0
            label = None
            label_angle = -pi / 2
            label_dist = 0.0
            label_color = "black"
            font = "sans-serif"
            label_size = 12.0
            # FIXME? mpl.rcParams["font.size"])
            position = {"func": self.layout.__getitem__}
            shape = "circle"
            size = 20.0
            width = None
            height = None
            zorder = 2

        return VisualVertexBuilder

    def draw(self, visual_vertex, vertex, point):
        if visual_vertex.size <= 0:
            return

        fig = self.fig

        marker_kwds = {}
        marker_kwds["x"] = [point[0]]
        marker_kwds["y"] = [point[1]]
        marker_kwds["marker"] = {
            "symbol": visual_vertex.shape,
            "size": visual_vertex.size,
            "color": format_rgba(visual_vertex.color),
            "line_color": format_rgba(visual_vertex.frame_color),
        }

        # if visual_vertex.label is not None:
        #    text_kwds['x'].append(point[0])
        #    text_kwds['y'].append(point[1])
        #    text_kwds['text'].append(visual_vertex.label)

        # Draw dots
        stroke = plotly.graph_objects.Scatter(
            mode="markers",
            **marker_kwds,
        )
        fig.add_trace(stroke)

    def draw_label(self, visual_vertex, point, **kwds):
        if visual_vertex.label is None:
            return

        fig = self.fig

        text_kwds = {}
        text_kwds["x"] = [point[0]]
        text_kwds["y"] = [point[1]]
        text_kwds["text"].append(visual_vertex.label)

        # TODO: add more options

        # Draw text labels
        stroke = plotly.graph_objects.Scatter(
            mode="text",
            **text_kwds,
        )
        fig.add_trace(stroke)



from contextlib import contextmanager
from typing import Iterator

__all__ = ("safe_locale",)


@contextmanager
def safe_locale() -> Iterator[None]:
    """Helper function that establishes a context that temporarily switches the
    current locale to use decimal dots when printing numbers.

    This can be used to establish an execution context within which it is safe
    to call functions that read or write graphs from/to the disk and ensure that
    they use decimal dots for interoperability with systems that are running
    with another locale.
    """
    from igraph._igraph import _enter_safelocale, _exit_safelocale

    locale = _enter_safelocale()
    try:
        yield
    finally:
        _exit_safelocale(locale)



"""
.. _tutorials-complement:

================
Complement
================

This example shows how to generate the `complement graph <https://en.wikipedia.org/wiki/Complement_graph>`_ of a graph (sometimes known as the anti-graph) using :meth:`igraph.GraphBase.complementer`.
"""
import igraph as ig
import matplotlib.pyplot as plt
import random

# %%
# First, we generate a random graph
random.seed(0)
g1 = ig.Graph.Erdos_Renyi(n=10, p=0.5)

# %%
# .. note::
#     We set the random seed to ensure the graph comes out exactly the same
#     each time in the gallery. You don't need to do that if you're exploring
#     really random graphs ;-)

# %%
# Then we generate the complement graph:
g2 = g1.complementer(loops=False)

# %%
# The union graph of the two is of course the full graph, i.e. a graph with
# edges connecting all vertices to all other vertices. Because we decided to
# ignore loops (aka self-edges) in the complementer, the full graph does not
# include loops either.
g_full = g1 | g2

# %%
# In case there was any doubt, the complement of the full graph is an
# empty graph, with the same vertices but no edges:
g_empty = g_full.complementer(loops=False)

# %%
# To demonstrate these concepts more clearly, here's a layout of each of the
# four graphs we discussed (input, complement, union/full, complement of
# union/empty):
fig, axs = plt.subplots(2, 2)
ig.plot(
    g1,
    target=axs[0, 0],
    layout="circle",
    vertex_color="black",
)
axs[0, 0].set_title('Original graph')
ig.plot(
    g2,
    target=axs[0, 1],
    layout="circle",
    vertex_color="black",
)
axs[0, 1].set_title('Complement graph')

ig.plot(
    g_full,
    target=axs[1, 0],
    layout="circle",
    vertex_color="black",
)
axs[1, 0].set_title('Union graph')
ig.plot(
    g_empty,
    target=axs[1, 1],
    layout="circle",
    vertex_color="black",
)
axs[1, 1].set_title('Complement of union graph')
plt.show()



# vim:ts=4 sw=4 sts=4:

import unittest

from igraph import Graph, Edge, EdgeSeq, VertexSeq

from .utils import is_pypy

try:
    import numpy as np
except ImportError:
    np = None


class EdgeTests(unittest.TestCase):
    def setUp(self):
        self.g = Graph.Full(10)

    def testHash(self):
        data = {}
        n = self.g.ecount()
        for i in range(n):
            code1 = hash(self.g.es[i])
            code2 = hash(self.g.es[i])
            self.assertEqual(code1, code2)
            data[self.g.es[i]] = i

        for i in range(n):
            self.assertEqual(i, data[self.g.es[i]])

    def testRichCompare(self):
        idxs = [2, 5, 9, 13, 42]
        g2 = Graph.Full(10)
        for i in idxs:
            for j in idxs:
                self.assertEqual(i == j, self.g.es[i] == self.g.es[j])
                self.assertEqual(i != j, self.g.es[i] != self.g.es[j])
                self.assertEqual(i < j, self.g.es[i] < self.g.es[j])
                self.assertEqual(i > j, self.g.es[i] > self.g.es[j])
                self.assertEqual(i <= j, self.g.es[i] <= self.g.es[j])
                self.assertEqual(i >= j, self.g.es[i] >= self.g.es[j])
                self.assertFalse(self.g.es[i] == g2.es[j])
                self.assertFalse(self.g.es[i] != g2.es[j])
                self.assertFalse(self.g.es[i] < g2.es[j])
                self.assertFalse(self.g.es[i] > g2.es[j])
                self.assertFalse(self.g.es[i] <= g2.es[j])
                self.assertFalse(self.g.es[i] >= g2.es[j])

        self.assertFalse(self.g.es[2] == self.g.vs[2])

    def testRepr(self):
        output = repr(self.g.es[0])
        self.assertEqual(output, "igraph.Edge(%r, 0, {})" % self.g)

        self.g.es["weight"] = list(range(10, 0, -1))
        output = repr(self.g.es[3])
        self.assertEqual(output, "igraph.Edge(%r, 3, {'weight': 7})" % self.g)

    def testUpdateAttributes(self):
        e = self.g.es[0]

        e.update_attributes(a=2)
        self.assertEqual(e["a"], 2)

        e.update_attributes([("a", 3), ("b", 4)], c=5, d=6)
        self.assertEqual(e.attributes(), {"a": 3, "b": 4, "c": 5, "d": 6})

        e.update_attributes({"b": 44, "c": 55})
        self.assertEqual(e.attributes(), {"a": 3, "b": 44, "c": 55, "d": 6})

    def testPhantomEdge(self):
        e = self.g.es[self.g.ecount() - 1]
        e.delete()

        # v is now a phantom edge; try to freak igraph out now :)
        self.assertRaises(ValueError, e.update_attributes, a=2)
        self.assertRaises(ValueError, e.__getitem__, "a")
        self.assertRaises(ValueError, e.__setitem__, "a", 4)
        self.assertRaises(ValueError, e.__delitem__, "a")
        self.assertRaises(ValueError, e.attributes)
        self.assertRaises(ValueError, getattr, e, "source")
        self.assertRaises(ValueError, getattr, e, "source_vertex")
        self.assertRaises(ValueError, getattr, e, "target")
        self.assertRaises(ValueError, getattr, e, "target_vertex")
        self.assertRaises(ValueError, getattr, e, "tuple")
        self.assertRaises(ValueError, getattr, e, "vertex_tuple")

    @unittest.skipIf(
        is_pypy, "skipped on PyPy because we do not have access to docstrings"
    )
    def testProxyMethods(self):
        g = Graph.GRG(10, 0.5)
        e = g.es[0]

        # - delete() is ignored because it mutates the graph
        ignore = "delete"
        ignore = set(ignore.split())

        # Methods not listed here are expected to return an int or a float
        return_types = {}

        for name in Edge.__dict__:
            if name in ignore:
                continue

            func = getattr(e, name)
            docstr = func.__doc__

            if not docstr.startswith("Proxy method"):
                continue

            result = func()
            self.assertEqual(
                getattr(g, name)(e.index),
                result,
                msg=("Edge.%s proxy method misbehaved" % name),
            )

            return_type = return_types.get(name, (int, float))
            self.assertTrue(
                isinstance(result, return_type),
                msg=("Edge.%s proxy method did not return %s" % (name, return_type)),
            )


class EdgeSeqTests(unittest.TestCase):
    def assert_edges_unique_in(self, es):
        pairs = sorted(e.tuple for e in es)
        self.assertEqual(pairs, sorted(set(pairs)))

    def setUp(self):
        self.g = Graph.Full(10)
        self.g.es["test"] = list(range(45))

    def testCreation(self):
        self.assertTrue(len(EdgeSeq(self.g)) == 45)
        self.assertTrue(len(EdgeSeq(self.g, 2)) == 1)
        self.assertTrue(len(EdgeSeq(self.g, [1, 2, 3])) == 3)
        self.assertTrue(EdgeSeq(self.g, [1, 2, 3]).indices == [1, 2, 3])
        self.assertRaises(ValueError, EdgeSeq, self.g, 112)
        self.assertRaises(ValueError, EdgeSeq, self.g, [112])
        self.assertTrue(self.g.es.graph == self.g)

    def testIndexing(self):
        n = self.g.ecount()
        for i in range(n):
            self.assertEqual(i, self.g.es[i].index)
            self.assertEqual(n - i - 1, self.g.es[-i - 1].index)
        self.assertRaises(IndexError, self.g.es.__getitem__, n)
        self.assertRaises(IndexError, self.g.es.__getitem__, -n - 1)
        self.assertRaises(TypeError, self.g.es.__getitem__, 1.5)

    @unittest.skipIf(np is None, "test case depends on NumPy")
    def testNumPyIndexing(self):
        assert np is not None

        n = self.g.ecount()
        for i in range(n):
            arr = np.array([i])
            self.assertEqual(i, self.g.es[arr[0]].index)

        arr = np.array([n])
        self.assertRaises(IndexError, self.g.es.__getitem__, arr[0])

        arr = np.array([-n - 1])
        self.assertRaises(IndexError, self.g.es.__getitem__, arr[0])

        arr = np.array([1.5])
        self.assertRaises(TypeError, self.g.es.__getitem__, arr[0])

        ind = [1, 3, 5, 8, 3, 2]
        arr = np.array(ind)
        self.assertEqual(ind, [edge.index for edge in self.g.es[arr.tolist()]])
        self.assertEqual(ind, [edge.index for edge in self.g.es[list(arr)]])

    def testPartialAttributeAssignment(self):
        only_even = self.g.es.select(lambda e: (e.index % 2 == 0))

        only_even["test"] = [0] * len(only_even)
        expected = [[0, i][i % 2] for i in range(self.g.ecount())]
        self.assertTrue(self.g.es["test"] == expected)

        only_even["test2"] = list(range(23))
        expected = [[i // 2, None][i % 2] for i in range(self.g.ecount())]
        self.assertTrue(self.g.es["test2"] == expected)

    def testSequenceReusing(self):
        if "test" in self.g.edge_attributes():
            del self.g.es["test"]

        self.g.es["test"] = ["A", "B", "C"]
        self.assertTrue(self.g.es["test"] == ["A", "B", "C"] * 15)
        self.g.es["test"] = "ABC"
        self.assertTrue(self.g.es["test"] == ["ABC"] * 45)

        only_even = self.g.es.select(lambda e: (e.index % 2 == 0))
        only_even["test"] = ["D", "E"]
        expected = ["D", "ABC", "E", "ABC"] * 12
        expected = expected[0:45]
        self.assertTrue(self.g.es["test"] == expected)
        del self.g.es["test"]
        only_even["test"] = ["D", "E"]
        expected = ["D", None, "E", None] * 12
        expected = expected[0:45]
        self.assertTrue(self.g.es["test"] == expected)

    def testAllSequence(self):
        self.assertTrue(len(self.g.es) == 45)
        self.assertTrue(self.g.es["test"] == list(range(45)))

    def testEmptySequence(self):
        empty_es = self.g.es.select(None)
        self.assertTrue(len(empty_es) == 0)
        self.assertRaises(IndexError, empty_es.__getitem__, 0)
        self.assertRaises(KeyError, empty_es.__getitem__, "nonexistent")
        self.assertTrue(empty_es["test"] == [])
        empty_es = self.g.es[[]]
        self.assertTrue(len(empty_es) == 0)
        empty_es = self.g.es[()]
        self.assertTrue(len(empty_es) == 0)

    def testCallableFilteringFind(self):
        edge = self.g.es.find(lambda e: (e.index % 2 == 1))
        self.assertTrue(edge.index == 1)
        self.assertRaises(IndexError, self.g.es.find, lambda e: (e.index % 2 == 3))

    def testCallableFilteringSelect(self):
        only_even = self.g.es.select(lambda e: (e.index % 2 == 0))
        self.assertTrue(len(only_even) == 23)
        self.assertRaises(KeyError, only_even.__getitem__, "nonexistent")
        self.assertTrue(only_even["test"] == [i * 2 for i in range(23)])

    def testChainedCallableFilteringSelect(self):
        only_div_six = self.g.es.select(
            lambda e: (e.index % 2 == 0), lambda e: (e.index % 3 == 0)
        )
        self.assertTrue(len(only_div_six) == 8)
        self.assertTrue(only_div_six["test"] == [0, 6, 12, 18, 24, 30, 36, 42])

        only_div_six = self.g.es.select(lambda e: (e.index % 2 == 0)).select(
            lambda e: (e.index % 3 == 0)
        )
        self.assertTrue(len(only_div_six) == 8)
        self.assertTrue(only_div_six["test"] == [0, 6, 12, 18, 24, 30, 36, 42])

    def testIntegerFilteringFind(self):
        self.assertEqual(self.g.es.find(3).index, 3)
        self.assertEqual(self.g.es.select(2, 3, 4, 2).find(3).index, 2)
        self.assertRaises(IndexError, self.g.es.find, 178)

    def testIntegerFilteringSelect(self):
        subset = self.g.es.select(2, 3, 4, 2)
        self.assertTrue(len(subset) == 4)
        self.assertTrue(subset["test"] == [2, 3, 4, 2])
        self.assertRaises(TypeError, self.g.es.select, 2, 3, 4, 2, None)

        subset = self.g.es[2, 3, 4, 2]
        self.assertTrue(len(subset) == 4)
        self.assertTrue(subset["test"] == [2, 3, 4, 2])

    def testIterableFilteringSelect(self):
        subset = self.g.es.select(list(range(5, 8)))
        self.assertTrue(len(subset) == 3)
        self.assertTrue(subset["test"] == [5, 6, 7])

    def testSliceFilteringSelect(self):
        subset = self.g.es.select(slice(5, 8))
        self.assertTrue(len(subset) == 3)
        self.assertTrue(subset["test"] == [5, 6, 7])
        subset = self.g.es[40:56:2]
        self.assertTrue(len(subset) == 3)
        self.assertTrue(subset["test"] == [40, 42, 44])

    def testKeywordFilteringSelect(self):
        g = Graph.Barabasi(1000, 2)
        g.es["betweenness"] = g.edge_betweenness()
        g.es["parity"] = [i % 2 for i in range(g.ecount())]
        self.assertTrue(len(g.es(betweenness_gt=10)) < 2000)
        self.assertTrue(len(g.es(betweenness_gt=10, parity=0)) < 2000)

    def testSourceTargetFiltering(self):
        g = Graph.Barabasi(1000, 2, directed=True)
        es1 = {e.source for e in g.es.select(_target_in=[2, 4])}
        es2 = {v1 for v1, v2 in g.get_edgelist() if v2 in [2, 4]}
        self.assertTrue(es1 == es2)

    def testWithinFiltering(self):
        g = Graph.Lattice([10, 10])
        vs = [0, 1, 2, 10, 11, 12, 20, 21, 22]
        vs2 = (0, 1, 10, 11)

        es1 = g.es.select(_within=vs)
        es2 = g.es.select(_within=VertexSeq(g, vs))

        for es in [es1, es2]:
            self.assertTrue(len(es) == 12)
            self.assertTrue(all(e.source in vs and e.target in vs for e in es))
            self.assert_edges_unique_in(es)

            es_filtered = es.select(_within=vs2)
            self.assertTrue(len(es_filtered) == 4)
            self.assertTrue(
                all(e.source in vs2 and e.target in vs2 for e in es_filtered)
            )
            self.assert_edges_unique_in(es_filtered)

    def testBetweenFiltering(self):
        g = Graph.Lattice([10, 10])
        vs1, vs2 = [10, 11, 12], [20, 21, 22]

        es1 = g.es.select(_between=(vs1, vs2))
        es2 = g.es.select(_between=(VertexSeq(g, vs1), VertexSeq(g, vs2)))

        for es in [es1, es2]:
            self.assertTrue(len(es) == 3)
            self.assertTrue(
                all(
                    (e.source in vs1 and e.target in vs2)
                    or (e.target in vs1 and e.source in vs2)
                    for e in es
                )
            )
            self.assert_edges_unique_in(es)

    def testIncidentFiltering(self):
        g = Graph.Lattice([10, 10], circular=False)
        vs = (0, 1, 10, 11)
        vs2 = (11, 0, 24)
        vs3 = sorted(set(vs).intersection(set(vs2)))

        es = g.es.select(_incident=vs)
        self.assertEqual(8, len(es))
        self.assertTrue(all((e.source in vs or e.target in vs) for e in es))
        self.assert_edges_unique_in(es)

        es_filtered = es.select(_incident=vs2)
        self.assertEqual(6, len(es_filtered))
        self.assertTrue(all((e.source in vs3 or e.target in vs3) for e in es_filtered))
        self.assert_edges_unique_in(es_filtered)

    def testIncidentFilteringDirected(self):
        # Test case from https://igraph.discourse.group/t/edge-select-using-incident-on-directed-graphs/1645
        g = Graph([(0, 1), (1, 2), (2, 3)], directed=True)

        vs = (1,)
        es = g.es.select(_incident=vs)
        self.assertEqual(2, len(es))
        self.assertTrue(all((e.source in vs or e.target in vs) for e in es))
        self.assert_edges_unique_in(es)

    def testIncidentFilteringByNames(self):
        g = Graph.Lattice([10, 10], circular=False)
        vs = (0, 1, 10, 11)
        g.vs[vs]["name"] = ["A", "B", "C", "D"]

        vs2 = (11, 0, 24)
        g.vs[24]["name"] = "X"

        vs3 = sorted(set(vs).intersection(set(vs2)))

        es = g.es.select(_incident=("A", "B", "C", "D"))
        self.assertEqual(8, len(es))
        self.assertTrue(all((e.source in vs or e.target in vs) for e in es))
        self.assert_edges_unique_in(es)

        es_filtered = es.select(_incident=("D", "A", "X"))
        self.assertEqual(6, len(es_filtered))
        self.assertTrue(all((e.source in vs3 or e.target in vs3) for e in es_filtered))
        self.assert_edges_unique_in(es_filtered)

        es_filtered = es_filtered.select(_from="A")
        self.assertEqual(2, len(es_filtered))
        self.assertTrue(all((e.source == 0 or e.target == 0) for e in es_filtered))
        self.assert_edges_unique_in(es_filtered)

    def testSourceAndTargetFilteringForUndirectedGraphs(self):
        g = Graph.Lattice([10, 10], circular=False)
        vs = (0, 1, 10, 11)
        vs2 = (11, 0, 24)
        vs3 = sorted(set(vs).intersection(set(vs2)))

        es = g.es.select(_from=vs)
        self.assertEqual(8, len(es))
        self.assertTrue(all((e.source in vs or e.target in vs) for e in es))
        self.assert_edges_unique_in(es)

        es_filtered = es.select(_to_in=vs2)
        self.assertEqual(6, len(es_filtered))
        self.assertTrue(all((e.source in vs3 or e.target in vs3) for e in es_filtered))
        self.assert_edges_unique_in(es_filtered)

        es_filtered = es_filtered.select(_from_eq=0)
        self.assertEqual(2, len(es_filtered))
        self.assertTrue(all((e.source == 0 or e.target == 0) for e in es_filtered))
        self.assert_edges_unique_in(es_filtered)

    def testIndexOutOfBoundsSelect(self):
        g = Graph.Full(3)
        self.assertRaises(ValueError, g.es.select, 4)
        self.assertRaises(ValueError, g.es.select, 4, 5)
        self.assertRaises(ValueError, g.es.select, (4, 5))
        self.assertRaises(ValueError, g.es.select, 2, -1)
        self.assertRaises(ValueError, g.es.select, (2, -1))
        self.assertRaises(ValueError, g.es.__getitem__, (0, 1000000))

    def testIndexAndKeywordFilteringFind(self):
        self.assertRaises(ValueError, self.g.es.find, 2, test=4)
        self.assertTrue(self.g.es.find(2, test=2) == self.g.es[2])

    def testGraphMethodProxying(self):
        idxs = [1, 3, 5, 7, 9]
        g = Graph.Barabasi(100)
        es = g.es(*idxs)
        ebs = g.edge_betweenness()
        self.assertEqual([ebs[i] for i in idxs], es.edge_betweenness())

        idxs = [1, 3]
        g = Graph([(0, 1), (1, 2), (2, 0), (1, 0), (2, 2)], directed=True)
        es = g.es(*idxs)
        mutual = g.is_mutual(es)
        self.assertEqual(mutual, es.is_mutual())
        for e, m in zip(es, mutual):
            self.assertEqual(e.is_mutual(), m)

        self.assertTrue(g.es[4].is_mutual())
        self.assertFalse(g.es[4].is_mutual(loops=False))

    def testIsAll(self):
        g = Graph.Full(5)
        self.assertTrue(g.es.is_all())
        self.assertFalse(g.es.select(1, 2, 3).is_all())
        self.assertFalse(g.es.select(_within=[1, 2, 3]).is_all())


def suite():
    edge_suite = unittest.defaultTestLoader.loadTestsFromTestCase(EdgeTests)
    es_suite = unittest.defaultTestLoader.loadTestsFromTestCase(EdgeSeqTests)
    return unittest.TestSuite([edge_suite, es_suite])


def test():
    runner = unittest.TextTestRunner()
    runner.run(suite())


if __name__ == "__main__":
    test()



"""
.. _tutorials-isomorphism:

===========
Isomorphism
===========

This example shows how to check for `isomorphism <https://en.wikipedia.org/wiki/Graph_isomorphism>`_ between small graphs using :meth:`igraph.GraphBase.isomorphic`.
"""
import igraph as ig
import matplotlib.pyplot as plt

# %%
# First we generate three different graphs:
g1 = ig.Graph([(0, 1), (0, 2), (0, 4), (1, 2), (1, 3), (2, 3), (2, 4), (3, 4)])
g2 = ig.Graph([(4, 2), (4, 3), (4, 0), (2, 3), (2, 1), (3, 1), (3, 0), (1, 0)])
g3 = ig.Graph([(4, 1), (4, 3), (4, 0), (2, 3), (2, 1), (3, 1), (3, 0), (1, 0)])

# %%
# To check whether they are isomorphic, we can use a simple method:
print("Are the graphs g1 and g2 isomorphic?")
print(g1.isomorphic(g2))
print("Are the graphs g1 and g3 isomorphic?")
print(g1.isomorphic(g3))
print("Are the graphs g2 and g3 isomorphic?")
print(g2.isomorphic(g3))

# Output:
# Are the graphs g1 and g2 isomorphic?
# True
# Are the graphs g1 and g3 isomorphic?
# False
# Are the graphs g2 and g3 isomorphic?
# False

# %%
# .. note::
#    `Graph isomorphism <https://en.wikipedia.org/wiki/Graph_isomorphism>`_ is an equivalence
#    relationship, i.e. if `g1 ~ g2` and `g2 ~ g3`, then automatically `g1 ~ g3`. Therefore,
#    we could have skipped the last check.

# %%
# We can plot the graphs to get an idea about the problem:
visual_style = {
    "vertex_color": "lightblue",
    "vertex_label": [0, 1, 2, 3, 4],
    "vertex_size": 25,
}

fig, axs = plt.subplots(1, 3)
ig.plot(
    g1,
    layout=g1.layout("circle"),
    target=axs[0],
    **visual_style,
)
ig.plot(
    g2,
    layout=g1.layout("circle"),
    target=axs[1],
    **visual_style,
)
ig.plot(
    g3,
    layout=g1.layout("circle"),
    target=axs[2],
    **visual_style,
)
fig.text(0.38, 0.5, '$\\simeq$' if g1.isomorphic(g2) else '$\\neq$', fontsize=15, ha='center', va='center')
fig.text(0.65, 0.5, '$\\simeq$' if g2.isomorphic(g3) else '$\\neq$', fontsize=15, ha='center', va='center')
plt.show()



#!/usr/bin/env python
"""
Runs all the doctests in the igraph module
"""

import doctest
import sys


def run_doctests():
    import igraph

    num_failed = 0
    modules = [
        igraph,
        igraph.clustering,
        igraph.cut,
        igraph.datatypes,
        igraph.drawing.utils,
        igraph.formula,
        igraph.statistics,
        igraph.utils,
    ]
    for module in modules:
        num_failed += doctest.testmod(module)[0]
    return num_failed == 0


if __name__ == "__main__":
    if run_doctests():
        sys.exit(0)
    else:
        sys.exit(1)



# -*- coding: utf-8 -*-
#
# igraph documentation build configuration file, created by
# sphinx-quickstart on Thu Jun 17 11:36:14 2010.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

from datetime import datetime

import sys
import os
import importlib
from pathlib import Path


# Check if we are inside readthedocs, the conf is quite different there
rtd_version = os.getenv("READTHEDOCS_VERSION", "")
rtd_version_type = os.getenv("READTHEDOCS_VERSION_TYPE", "unknown")

# Utility functions
# NOTE: these could be improved, esp by importing igraph, but that
# currently generates a conflict with pydoctor. It is funny because pydoctor's
# docs indeed import itself... perhaps there's a decent way to solve this.
def get_root_dir():
    """Get project root folder"""
    return str(Path(".").absolute().parent.parent)


def get_igraphdir():
    """Get igraph folder"""
    return Path(importlib.util.find_spec("igraph").origin).parent


def get_igraph_version():
    """Get igraph version"""
    if rtd_version and rtd_version_type == "tag":
        return rtd_version

    version_file = get_igraphdir() / "version.py"
    with open(version_file, "rt") as f:
        version_info = f.readline().rstrip("\n").split("=")[1].strip()[1:-1].split(", ")
    version = ".".join(version_info)

    return version


# -- General configuration -----------------------------------------------------

_igraph_dir = str(get_igraphdir())
_igraph_version = get_igraph_version()

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.intersphinx",
    "sphinx_gallery.gen_gallery",
    #'sphinx_panels',
    "pydoctor.sphinx_ext.build_apidocs",
]

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.append(os.path.abspath('.'))

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "igraph"
copyright = "2010-{0}, The igraph development team".format(datetime.now().year)

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = _igraph_version
# The full version, including alpha/beta/rc tags.
release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["include/*.rst"]

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------

# Inspired by pydoctor's RTD page itself
# https://github.com/twisted/pydoctor/blob/master/docs/source/conf.py
html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
html_css_files = ["custom.css"]

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = "_static/logo-black.svg"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# If nonempty, this is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = ''

# Output file base name for HTML help builder.
htmlhelp_basename = "igraphdoc"

# Integration with Read the Docs since RTD is not manipulating the Sphinx
# config files on its own any more.
# This is according to:
# https://about.readthedocs.com/blog/2024/07/addons-by-default/
html_baseurl = os.environ.get("READTHEDOCS_CANONICAL_URL", "")
html_context = {}
if os.environ.get("READTHEDOCS", "") == "True":
    html_context["READTHEDOCS"] = True

# -- Options for pydoctor ------------------------------------------------------


def get_pydoctor_html_outputdir(pydoctor_url_path):
    """Get HTML output dir for pydoctor"""
    # NOTE: obviously this is a little tricky, but it does work for both
    # the sphinx-build script and the python -m sphinx module calls. It works
    # locally, on github pages, and on RTD.
    return str(Path(sys.argv[-1]) / pydoctor_url_path.strip("/"))


# API docs relative to the rest of the docs, needed for pydoctor to play nicely
# with intersphinx (https://pypi.org/project/pydoctor/#pydoctor-21-2-0)
# NOTE: As of 2022 AD, pydoctor requires this to be a subfolder of the docs.
pydoctor_url_path = "api/"

pydoctor_args = [
    "--project-name=igraph",
    "--project-version=" + version,
    "--project-url=https://python.igraph.org",
    "--introspect-c-modules",
    "--docformat=epytext",
    "--intersphinx=https://docs.python.org/3/objects.inv",
    "--html-output=" + get_pydoctor_html_outputdir(pydoctor_url_path),
    "--html-viewsource-base=https://github.com/igraph/python-igraph/tree/main/src/igraph",
    "--project-base-dir=" + _igraph_dir,
    "--template-dir=" + get_root_dir() + "/doc/source/_pydoctor_templates",
    "--theme=readthedocs",
    _igraph_dir,
]
pydoctor_url_path = "/en/{rtd_version}/api"


# -- Options for sphinx-gallery ------------------------------------------------

sphinx_gallery_conf = {
    "examples_dirs": "../examples_sphinx-gallery",  # path to your example scripts
    "gallery_dirs": "tutorials",  # path to where to save gallery generated output
    "filename_pattern": "/",
    "matplotlib_animations": True,
    "remove_config_comments": True,
}

# -- Options for LaTeX output --------------------------------------------------

# The paper size ('letter' or 'a4').
# latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
# latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    (
        "index",
        "igraph.tex",
        "igraph Documentation",
        "The igraph development team",
        "manual",
    ),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Additional stuff for the LaTeX preamble.
# latex_preamble = ''

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ("index", "igraph", "igraph Documentation", ["The igraph development team"], 1)
]


# -- Options for Epub output ---------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = "igraph"
epub_author = "The igraph development team"
epub_publisher = "The igraph development team"
epub_copyright = "2010-2022, The igraph development team"

# The language of the text. It defaults to the language option
# or en if the language is not set.
# epub_language = ''

# The scheme of the identifier. Typical schemes are ISBN or URL.
# epub_scheme = ''

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
# epub_identifier = ''

# A unique identification for the text.
# epub_uid = ''

# HTML files that should be inserted before the pages created by sphinx.
# The format is a list of tuples containing the path and title.
# epub_pre_files = []

# HTML files shat should be inserted after the pages created by sphinx.
# The format is a list of tuples containing the path and title.
# epub_post_files = []

# A list of files that should not be packed into the epub file.
# epub_exclude_files = []

# The depth of the table of contents in toc.ncx.
# epub_tocdepth = 3


# -- Intersphinx ------------------------------------------------

intersphinx_mapping = {
    "numpy": ("https://numpy.org/doc/stable/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
    "matplotlib": ("https://matplotlib.org/stable", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "networkx": ("https://networkx.org/documentation/stable/", None),
}



from igraph.drawing.utils import calculate_corner_radii
from igraph.utils import consecutive_pairs

from .utils import find_plotly, format_path_step

__all__ = ("PlotlyPolygonDrawer",)

plotly = find_plotly()


class PlotlyPolygonDrawer:
    """Class that is used to draw polygons in matplotlib.

    The corner points of the polygon can be set by the C{points}
    property of the drawer, or passed at construction time. Most
    drawing methods in this class also have an extra C{points}
    argument that can be used to override the set of points in the
    C{points} property."""

    def __init__(self, fig):
        """Constructs a new polygon drawer that draws on the given
        Matplotlib axes.

        @param  fig: the plotly Figure to draw on
        """
        self.context = fig

    def draw(self, points, corner_radius=0, **kwds):
        """Draws a polygon to the associated axes.

        @param points: the coordinates of the corners of the polygon,
          in clockwise or counter-clockwise order, or C{None} if we are
          about to use the C{points} property of the class.
        @param corner_radius: if zero, an ordinary polygon will be drawn.
          If positive, the corners of the polygon will be rounded with
          the given radius.
        """
        if len(points) < 2:
            # Well, a polygon must have at least two corner points
            return

        fig = self.context
        if corner_radius <= 0:
            # No rounded corners, this is simple
            # We need to repeat the initial point to get a closed shape
            x = [p[0] for p in points] + [points[0][0]]
            y = [p[1] for p in points] + [points[0][1]]
            kwds["mode"] = kwds.get("mode", "line")
            stroke = plotly.graph_objects.Scatter(
                x=x,
                y=y,
                **kwds,
            )
            fig.add_trace(stroke)

        # Rounded corners. First, we will take each side of the
        # polygon and find what the corner radius should be on
        # each corner. If the side is longer than 2r (where r is
        # equal to corner_radius), the radius allowed by that side
        # is r; if the side is shorter, the radius is the length
        # of the side / 2. For each corner, the final corner radius
        # is the smaller of the radii on the two sides adjacent to
        # the corner.
        corner_radii = calculate_corner_radii(points, corner_radius)

        # Okay, move to the last corner, adjusted by corner_radii[-1]
        # towards the first corner
        path = []
        path.append(
            format_path_step(
                "M",
                [points[-1].towards(points[0], corner_radii[-1])],
            )
        )

        # Now, for each point in points, draw a line towards the
        # corner, stopping before it in a distance of corner_radii[idx],
        # then draw the corner
        u = points[-1]
        for idx, (v, w) in enumerate(consecutive_pairs(points, True)):
            radius = corner_radii[idx]
            path.append(
                format_path_step(
                    "L",
                    [v.towards(u, radius)],
                )
            )

            aux1 = v.towards(u, radius / 2)
            aux2 = v.towards(w, radius / 2)

            path.append(
                format_path_step(
                    "C",
                    [aux1, aux2, v.towards(w, corner_radii[idx])],
                )
            )
            u = v

        # Close path
        path = "".join(path).strip(" ") + " Z"
        stroke = dict(
            type="path",
            path=path,
            **kwds,
        )
        fig.update_layout(
            shapes=[stroke],
        )



"""
.. _tutorials-configuration:

======================
Configuration Instance
======================

This example shows how to use igraph's :class:`configuration instance <igraph.Configuration>` to set default igraph settings. This is useful for setting global settings so that they don't need to be explicitly stated at the beginning of every igraph project you work on.
"""
import igraph as ig
import matplotlib.pyplot as plt
import random

# %%
# First we define the default plotting backend, layout, and color palette.
ig.config["plotting.backend"] = "matplotlib"
ig.config["plotting.layout"] = "fruchterman_reingold"
ig.config["plotting.palette"] = "rainbow"

# %%
# Then, we save them. By default, ``ig.config.save()`` will save files to
# ``~/.igraphrc`` on Linux and Max OS X systems, or in
# ``%USERPROFILE%\.igraphrc`` for Windows systems:
ig.config.save()

# %%
# The code above only needs to be run once (to store the new config options
# into the ``.igraphrc`` file). Whenever you use igraph and this file exists,
# igraph will read its content and use those options as defaults. For
# example, let's create and plot a new graph to demonstrate:
random.seed(1)
g = ig.Graph.Barabasi(n=100, m=1)

# %%
# We now calculate a color value between 0-200 for all nodes, for instance by
# computing the vertex betweenness:
betweenness = g.betweenness()
colors = [int(i * 200 / max(betweenness)) for i in betweenness]

# %%
# Finally, we can plot the graph. You will notice that even though we did not
# create a dedicated figure and axes, matplotlib is now used by default:
ig.plot(g, vertex_color=colors, vertex_size=15, edge_width=0.3)
plt.show()

# %%
# The full list of config settings can be found at
# :class:`igraph.Configuration`.
#
# .. note::
#
#    You can have multiple config files: specify each location via
#    ``ig.config.save("./path/to/config/file")``. To load a specific config,
#    import igraph and then call ``ig.config.load("./path/to/config/file")``
#
#
# .. note::
#
#     To use a consistent style between individual plots (e.g. vertex sizes,
#     colors, layout etc.) check out :ref:`tutorials-visual-style`.



from math import pi
from typing import Tuple, Union

from igraph.drawing.baseclasses import AbstractDrawer
from igraph.drawing.utils import BoundingBox

__all__ = ("AbstractCairoDrawer",)


class AbstractCairoDrawer(AbstractDrawer):
    """Abstract class that serves as a base class for anything that
    draws on a Cairo context within a given bounding box.

    A subclass of L{AbstractCairoDrawer} is guaranteed to have an
    attribute named C{context} that represents the Cairo context
    to draw on, and an attribute named C{bbox} for the L{BoundingBox}
    of the drawing area.
    """

    _bbox: BoundingBox

    def __init__(self, context, bbox: BoundingBox or None):
        """Constructs the drawer and associates it to the given
        Cairo context and the given L{BoundingBox}.

        @param context: the context on which we will draw
        @param bbox:    the bounding box within which we will draw.
                        Can be anything accepted by the constructor
                        of L{BoundingBox} (i.e., a 2-tuple, a 4-tuple
                        or a L{BoundingBox} object).
        """
        self.context = context
        self._bbox = None  # type: ignore
        # can be set at drawing time
        if bbox is not None:
            self.bbox = bbox

    @property
    def bbox(self) -> BoundingBox:
        """The bounding box of the drawing area where this drawer will
        draw."""
        return self._bbox

    @bbox.setter
    def bbox(self, bbox):
        """Sets the bounding box of the drawing area where this drawer
        will draw."""
        if not isinstance(bbox, BoundingBox):
            self._bbox = BoundingBox(bbox)
        else:
            self._bbox = bbox

    def _mark_point(
        self,
        x: float,
        y: float,
        color: Union[int, Tuple[float, ...]] = 0,
        size: float = 4,
    ) -> None:
        """Marks the given point with a small circle on the canvas.
        Used primarily for debugging purposes.

        @param x: the X coordinate of the point to mark
        @param y: the Y coordinate of the point to mark
        @param color: the color of the marker. It can be a
          3-tuple (RGB components, alpha=0.5), a 4-tuple
          (RGBA components) or an index where zero means red, 1 means
          green, 2 means blue and so on.
        @param size: the diameter of the marker.
        """
        if isinstance(color, int):
            colors = [(1, 0, 0), (0, 1, 0), (0, 0, 1), (1, 1, 0), (0, 1, 1), (1, 0, 1)]
            color_tuple = colors[color % len(colors)]
        elif len(color) == 3:
            color_tuple = color + (0.5,)
        else:
            color_tuple = color

        ctx = self.context
        ctx.save()
        ctx.set_source_rgba(*color_tuple)
        ctx.arc(x, y, size / 2.0, 0, 2 * pi)
        ctx.fill()
        ctx.restore()



import unittest
from igraph import Graph


class BipartiteTests(unittest.TestCase):
    def testCreateBipartite(self):
        g = Graph.Bipartite([0, 1] * 5, [(0, 1), (2, 3), (4, 5), (6, 7), (8, 9)])
        self.assertTrue(
            g.vcount() == 10 and g.ecount() == 5 and g.is_directed() is False
        )
        self.assertTrue(g.is_bipartite())
        self.assertTrue(g.vs["type"] == [False, True] * 5)

    def testFullBipartite(self):
        g = Graph.Full_Bipartite(10, 5)
        self.assertTrue(
            g.vcount() == 15 and g.ecount() == 50 and g.is_directed() is False
        )
        expected = sorted([(i, j) for i in range(10) for j in range(10, 15)])
        self.assertTrue(sorted(g.get_edgelist()) == expected)
        self.assertTrue(g.vs["type"] == [False] * 10 + [True] * 5)

        g = Graph.Full_Bipartite(10, 5, directed=True, mode="out")
        self.assertTrue(
            g.vcount() == 15 and g.ecount() == 50 and g.is_directed() is True
        )
        self.assertTrue(sorted(g.get_edgelist()) == expected)
        self.assertTrue(g.vs["type"] == [False] * 10 + [True] * 5)

        g = Graph.Full_Bipartite(10, 5, directed=True, mode="in")
        self.assertTrue(
            g.vcount() == 15 and g.ecount() == 50 and g.is_directed() is True
        )
        self.assertTrue(
            sorted(g.get_edgelist()) == sorted([(i, j) for j, i in expected])
        )
        self.assertTrue(g.vs["type"] == [False] * 10 + [True] * 5)

        g = Graph.Full_Bipartite(10, 5, directed=True)
        self.assertTrue(
            g.vcount() == 15 and g.ecount() == 100 and g.is_directed() is True
        )
        expected.extend([(j, i) for i, j in expected])
        expected.sort()
        self.assertTrue(sorted(g.get_edgelist()) == expected)
        self.assertTrue(g.vs["type"] == [False] * 10 + [True] * 5)

    def testBiadjacency(self):
        g = Graph.Biadjacency([[0, 1, 1], [1, 2, 0]])
        self.assertTrue(all((g.vcount() == 5, g.ecount() == 4, not g.is_directed())))
        self.assertListEqual(g.vs["type"], [False] * 2 + [True] * 3)
        self.assertListEqual(sorted(g.get_edgelist()), [(0, 3), (0, 4), (1, 2), (1, 3)])

        g = Graph.Biadjacency([[0, 1, 1], [1, 2, 0]], multiple=True)
        self.assertTrue(all((g.vcount() == 5, g.ecount() == 5, not g.is_directed())))
        self.assertListEqual(g.vs["type"], [False] * 2 + [True] * 3)
        self.assertListEqual(
            sorted(g.get_edgelist()), [(0, 3), (0, 4), (1, 2), (1, 3), (1, 3)]
        )

        g = Graph.Biadjacency([[0, 1, 1], [1, 2, 0]], directed=True)
        self.assertTrue(all((g.vcount() == 5, g.ecount() == 4, g.is_directed())))
        self.assertListEqual(g.vs["type"], [False] * 2 + [True] * 3)
        self.assertListEqual(sorted(g.get_edgelist()), [(0, 3), (0, 4), (1, 2), (1, 3)])

        g = Graph.Biadjacency([[0, 1, 1], [1, 2, 0]], directed=True, mode="in")
        self.assertTrue(all((g.vcount() == 5, g.ecount() == 4, g.is_directed())))
        self.assertListEqual(g.vs["type"], [False] * 2 + [True] * 3)
        self.assertListEqual(sorted(g.get_edgelist()), [(2, 1), (3, 0), (3, 1), (4, 0)])

        # Create a weighted Graph
        g = Graph.Biadjacency([[0, 1, 1], [1, 2, 0]], weighted=True)
        self.assertTrue(
            all(
                (g.vcount() == 5, g.ecount() == 4, not g.is_directed(), g.is_weighted())
            )
        )
        self.assertListEqual(g.vs["type"], [False] * 2 + [True] * 3)
        self.assertListEqual(g.es["weight"], [1, 1, 1, 2])
        self.assertListEqual(sorted(g.get_edgelist()), [(0, 3), (0, 4), (1, 2), (1, 3)])

        # Graph is not weighted when weighted=`str`
        g = Graph.Biadjacency([[0, 1, 1], [1, 2, 0]], weighted="some_attr_name")
        self.assertTrue(
            all(
                (
                    g.vcount() == 5,
                    g.ecount() == 4,
                    not g.is_directed(),
                    not g.is_weighted(),
                )
            )
        )
        self.assertListEqual(g.vs["type"], [False] * 2 + [True] * 3)
        self.assertListEqual(g.es["some_attr_name"], [1, 1, 1, 2])
        self.assertListEqual(sorted(g.get_edgelist()), [(0, 3), (0, 4), (1, 2), (1, 3)])

        # Graph is not weighted when weighted=""
        g = Graph.Biadjacency([[0, 1, 1], [1, 2, 0]], weighted="")
        self.assertTrue(
            all(
                (
                    g.vcount() == 5,
                    g.ecount() == 4,
                    not g.is_directed(),
                    not g.is_weighted(),
                )
            )
        )
        self.assertListEqual(g.vs["type"], [False] * 2 + [True] * 3)
        self.assertListEqual(g.es[""], [1, 1, 1, 2])
        self.assertListEqual(sorted(g.get_edgelist()), [(0, 3), (0, 4), (1, 2), (1, 3)])

        # Should work when directed=True and mode=out with weighted
        g = Graph.Biadjacency([[0, 1, 1], [1, 2, 0]], directed=True, weighted=True)
        self.assertTrue(
            all((g.vcount() == 5, g.ecount() == 4, g.is_directed(), g.is_weighted()))
        )
        self.assertListEqual(g.vs["type"], [False] * 2 + [True] * 3)
        self.assertListEqual(g.es["weight"], [1, 1, 1, 2])
        self.assertListEqual(sorted(g.get_edgelist()), [(0, 3), (0, 4), (1, 2), (1, 3)])

        # Should work when directed=True and mode=in with weighted
        g = Graph.Biadjacency(
            [[0, 1, 1], [1, 2, 0]], directed=True, mode="in", weighted=True
        )
        self.assertTrue(
            all((g.vcount() == 5, g.ecount() == 4, g.is_directed(), g.is_weighted()))
        )
        self.assertListEqual(g.vs["type"], [False] * 2 + [True] * 3)
        self.assertListEqual(g.es["weight"], [1, 1, 1, 2])
        self.assertListEqual(sorted(g.get_edgelist()), [(2, 1), (3, 0), (3, 1), (4, 0)])

        # Should work when directed=True and mode=all with weighted
        g = Graph.Biadjacency(
            [[0, 1, 1], [1, 2, 0]], directed=True, mode="all", weighted=True
        )
        self.assertTrue(
            all((g.vcount() == 5, g.ecount() == 8, g.is_directed(), g.is_weighted()))
        )
        self.assertListEqual(g.vs["type"], [False] * 2 + [True] * 3)
        self.assertListEqual(g.es["weight"], [1, 1, 1, 1, 1, 1, 2, 2])
        self.assertListEqual(
            sorted(g.get_edgelist()),
            [(0, 3), (0, 4), (1, 2), (1, 3), (2, 1), (3, 0), (3, 1), (4, 0)],
        )

    def testBiadjacencyError(self):
        msg = "arguments weighted and multiple can not co-exist"
        with self.assertRaises(ValueError) as e:
            Graph.Biadjacency([[0, 1, 1], [1, 2, 0]], multiple=True, weighted=True)
        self.assertIn(msg, e.exception.args)

        with self.assertRaises(ValueError) as e:
            Graph.Biadjacency([[0, 1, 1], [1, 2, 0]], multiple=True, weighted="string")
        self.assertIn(msg, e.exception.args)

        with self.assertRaises(ValueError) as e:
            Graph.Biadjacency([[0, 1, 1], [1, 2, 0]], multiple=True, weighted="")
        self.assertIn(msg, e.exception.args)

    def testGetBiadjacency(self):
        mat = [[0, 1, 1], [1, 1, 0]]
        v1, v2 = [0, 1], [2, 3, 4]
        g = Graph.Biadjacency(mat)
        self.assertTrue(g.get_biadjacency() == (mat, v1, v2))
        g.vs["type2"] = g.vs["type"]
        self.assertTrue(g.get_biadjacency("type2") == (mat, v1, v2))
        self.assertTrue(g.get_biadjacency(g.vs["type2"]) == (mat, v1, v2))

    def testBipartiteProjection(self):
        g = Graph.Full_Bipartite(10, 5)

        g1, g2 = g.bipartite_projection()
        self.assertTrue(g1.is_complete())
        self.assertTrue(g1.isomorphic(Graph.Full(10)))
        self.assertTrue(g2.is_complete())
        self.assertTrue(g2.isomorphic(Graph.Full(5)))
        self.assertTrue(g.bipartite_projection(which=0).isomorphic(g1))
        self.assertTrue(g.bipartite_projection(which=1).isomorphic(g2))
        self.assertTrue(g.bipartite_projection(which=False).isomorphic(g1))
        self.assertTrue(g.bipartite_projection(which=True).isomorphic(g2))
        self.assertTrue(g1.es["weight"] == [5] * 45)
        self.assertTrue(g2.es["weight"] == [10] * 10)
        self.assertTrue(g.bipartite_projection_size() == (10, 45, 5, 10))

        g1, g2 = g.bipartite_projection(probe1=10)
        self.assertTrue(g1.is_complete())
        self.assertTrue(g1.isomorphic(Graph.Full(5)))
        self.assertTrue(g2.is_complete())
        self.assertTrue(g2.isomorphic(Graph.Full(10)))
        self.assertTrue(g.bipartite_projection(which=0).isomorphic(g2))
        self.assertTrue(g.bipartite_projection(which=1).isomorphic(g1))
        self.assertTrue(g.bipartite_projection(which=False).isomorphic(g2))
        self.assertTrue(g.bipartite_projection(which=True).isomorphic(g1))

        g1, g2 = g.bipartite_projection(multiplicity=False)
        self.assertTrue(g1.is_complete())
        self.assertTrue(g1.isomorphic(Graph.Full(10)))
        self.assertTrue(g2.is_complete())
        self.assertTrue(g2.isomorphic(Graph.Full(5)))
        self.assertTrue(g.bipartite_projection(which=0).isomorphic(g1))
        self.assertTrue(g.bipartite_projection(which=1).isomorphic(g2))
        self.assertTrue(g.bipartite_projection(which=False).isomorphic(g1))
        self.assertTrue(g.bipartite_projection(which=True).isomorphic(g2))
        self.assertTrue("weight" not in g1.edge_attributes())
        self.assertTrue("weight" not in g2.edge_attributes())

    def testIsBipartite(self):
        g = Graph.Star(10)
        self.assertTrue(g.is_bipartite() is True)
        self.assertTrue(g.is_bipartite(True) == (True, [False] + [True] * 9))
        g = Graph.Tree(100, 3)
        self.assertTrue(g.is_bipartite() is True)
        g = Graph.Ring(9)
        self.assertTrue(g.is_bipartite() is False)
        self.assertTrue(g.is_bipartite(True) == (False, None))
        g = Graph.Ring(10)
        self.assertTrue(g.is_bipartite() is True)
        g += (2, 0)
        self.assertTrue(g.is_bipartite(True) == (False, None))


def suite():
    bipartite_suite = unittest.defaultTestLoader.loadTestsFromTestCase(BipartiteTests)
    return unittest.TestSuite([bipartite_suite])


def test():
    runner = unittest.TextTestRunner()
    runner.run(suite())


if __name__ == "__main__":
    test()
